import os
import pytest
import torch, math
import triton, pickle
import triton.language as tl


@triton.jit
def _attn_fwd_inner(acc, l_i, m_i, q, q_scale, 
                    K_ptrs, K_scale_ptr, V_ptrs,  
                    start_m, qk_scale,  
                    BLOCK_M: tl.constexpr, HEAD_DIM: tl.constexpr, BLOCK_N: tl.constexpr,  
                    STAGE: tl.constexpr, offs_m: tl.constexpr, offs_n: tl.constexpr,  
                    N_CTX: tl.constexpr):
    lo, hi = 0, N_CTX
    for start_n in range(lo, hi, BLOCK_N):
        start_n = tl.multiple_of(start_n, BLOCK_N)
        k_mask = offs_n[None, :] < (N_CTX - start_n)  
        k = tl.load(K_ptrs, mask = k_mask)
        k_scale = tl.load(K_scale_ptr)
        qk = tl.dot(q, k).to(tl.float32) * q_scale * k_scale
        m_ij = tl.maximum(m_i, tl.max(qk, 1) * qk_scale)
        qk = qk * qk_scale - m_ij[:, None]
        p = tl.math.exp2(qk)
        l_ij = tl.sum(p, 1)
        alpha = tl.math.exp2(m_i - m_ij)
        l_i = l_i * alpha + l_ij
        acc = acc * alpha[:, None]
        v = tl.load(V_ptrs, mask = offs_n[:, None] < (N_CTX - start_n))
        p = p.to(tl.float16)
        
        acc += tl.dot(p, v, out_dtype=tl.float16)  
        m_i = m_ij
        K_ptrs += BLOCK_N * HEAD_DIM
        K_scale_ptr += 1
        V_ptrs += BLOCK_N * HEAD_DIM
    return acc, l_i

@triton.jit
def _attn_fwd(Q, K, V, Q_scale, K_scale, sm_scale, M, Out,  
              stride_qz, stride_qh, stride_qm, stride_qk,  
              stride_kz, stride_kh, stride_kn, stride_kk,  
              stride_vz, stride_vh, stride_vk, stride_vn,  
              stride_oz, stride_oh, stride_om, stride_on,  
              Z, H, N_CTX,  
              HEAD_DIM: tl.constexpr,  
              BLOCK_M: tl.constexpr,  
              BLOCK_N: tl.constexpr,  
              STAGE: tl.constexpr  
              ):
    start_m = tl.program_id(0)
    off_hz = tl.program_id(1)
    
    off_z = off_hz // H
    off_h = off_hz % H
    qvk_offset = off_z.to(tl.int64) * stride_qz + off_h.to(tl.int64) * stride_qh
    vk_offset = qvk_offset // stride_qm
    q_scale_offset = off_hz * tl.cdiv(N_CTX, BLOCK_M)
    k_scale_offset = off_hz * tl.cdiv(N_CTX, BLOCK_N) 
    

    offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
    offs_n = tl.arange(0, BLOCK_N)
    offs_k = tl.arange(0, HEAD_DIM)
    Q_ptrs = Q + qvk_offset + offs_m[:, None] * stride_qm + offs_k[None, :] * stride_qk
    Q_scale_ptr = Q_scale + q_scale_offset + start_m
    K_ptrs = K + qvk_offset + offs_k[:, None] + offs_n[None, :] * stride_kn
    K_scale_ptr = K_scale + k_scale_offset
    V_ptrs = V + qvk_offset + offs_n[:, None] * stride_qm + offs_k[None, :] * stride_qk
    O_block_ptr = Out + qvk_offset + offs_m[:, None] * stride_qm + offs_k[None, :] * stride_qk
    m_i = tl.zeros([BLOCK_M], dtype=tl.float32) - float("inf")
    l_i = tl.zeros([BLOCK_M], dtype=tl.float32) + 1.0
    acc = tl.zeros([BLOCK_M, HEAD_DIM], dtype=tl.float32)
    qk_scale = sm_scale
    qk_scale *= 1.44269504 
    q = tl.load(Q_ptrs, mask = offs_m[:, None] < N_CTX)
    q_scale = tl.load(Q_scale_ptr)
    acc, l_i = _attn_fwd_inner(acc, l_i, m_i, q, q_scale, K_ptrs, K_scale_ptr, V_ptrs,  
                                    start_m, qk_scale,  
                                    BLOCK_M, HEAD_DIM, BLOCK_N,  
                                    4 - STAGE, offs_m, offs_n, N_CTX 
                                    )
    acc = acc / l_i[:, None]
    tl.store(O_block_ptr, acc.to(Out.type.element_ty), mask = (offs_m[:, None] < N_CTX))


def per_block_int8(tensor, BLOCK_M=128):
    num_blocks =  (tensor.shape[2] // BLOCK_M) + (1 if tensor.shape[2] % BLOCK_M != 0 else 0)
    processed_tensors = []
    scales = []
    for i in range(num_blocks):
        start_idx = i * BLOCK_M
        end_idx = min(start_idx + BLOCK_M, tensor.shape[2])
        block_tensor = tensor[:, :, start_idx:end_idx, :]
        max_over_h, _ = torch.max(torch.abs(block_tensor), dim=2, keepdim=True)
        max_over_hw, _ = torch.max(max_over_h, dim=3, keepdim=True)
        scale = max_over_hw.to(torch.float32) / 127
        scaled_tensor = (block_tensor / scale).round().to(torch.int8).to(torch.float32)
        processed_tensors.append(scaled_tensor)
        scales.append(scale)
    final_tensor = torch.cat(processed_tensors, dim=2)
    final_scales = torch.cat(scales, dim=2)
    return final_tensor, final_scales


class _attention(torch.autograd.Function):
    @staticmethod
    def forward(ctx, q, k, v, q_scale, k_scale, causal, sm_scale):
        BLOCK_M = 128
        BLOCK_N = 64
        HEAD_DIM_Q, HEAD_DIM_K = q.shape[-1], k.shape[-1]
        HEAD_DIM_V = v.shape[-1]
        assert HEAD_DIM_Q == HEAD_DIM_K and HEAD_DIM_K == HEAD_DIM_V
        assert HEAD_DIM_K in {16, 32, 64, 128, 256}
        o = torch.empty_like(q, dtype=torch.float16)
        stage = 3 if causal else 1
        extra_kern_args = {}

    
        grid = (triton.cdiv(q.shape[2], BLOCK_M), q.shape[0] * q.shape[1], 1)
        M = torch.empty((q.shape[0], q.shape[1], q.shape[2]), device=q.device, dtype=torch.float32)
        _attn_fwd[grid](
            q, k, v, q_scale, k_scale, sm_scale, M, o,  
            q.stride(0), q.stride(1), q.stride(2), q.stride(3),  
            k.stride(0), k.stride(1), k.stride(2), k.stride(3),  
            v.stride(0), v.stride(1), v.stride(2), v.stride(3),  
            o.stride(0), o.stride(1), o.stride(2), o.stride(3),  
            q.shape[0], q.shape[1],  
            N_CTX=q.shape[2],  
            BLOCK_M=BLOCK_M, BLOCK_N=BLOCK_N, HEAD_DIM=HEAD_DIM_K,  
            STAGE=stage,  
            num_warps=4,  
            num_stages=3,  
            **extra_kern_args)
        return o

attention = _attention.apply


import torch.nn.functional as F
def precision_metric(quant_o, fa2_o): 
        x, xx = quant_o.float(), fa2_o.float() 
        sim = F.cosine_similarity(x.reshape(1, -1), xx.reshape(1, -1)).item()
        l1 =   ( (x - xx).abs().sum() / xx.abs().sum() ).item()
        rmse = torch.sqrt(torch.mean((x -xx) ** 2)).item()
        print(f'Cossim: {sim:.6f}, L1: {l1:.6f}, RMSE:{rmse:.6f}\n')



def test_op():
    q = (torch.randn((1, 4, 4096, 64), dtype=torch.float16, device="cuda")).to(torch.float16)
    k = (torch.randn((1, 4, 4096, 64), dtype=torch.float16, device="cuda")).to(torch.float16)
    v = (torch.randn((1, 4, 4096, 64), dtype=torch.float16, device="cuda")).to(torch.float16)

    torch_stand = torch.nn.functional.scaled_dot_product_attention(q, k, v)
    k -= k.mean(dim=-2, keepdim=True)
    
    sm_scale = 1 / math.sqrt(q.size(-1))
    q_int8, q_scale = per_block_int8(q, BLOCK_M=128)
    k_int8, k_scale = per_block_int8(k, BLOCK_M=64)
    q_int8 = q_int8.contiguous()
    k_int8 = k_int8.contiguous()
    q_int8 = q_int8.to(torch.int8)
    k_int8 = k_int8.to(torch.int8)
    tri_out = attention(q_int8, k_int8, v, q_scale, k_scale, False, sm_scale).half()
    precision_metric(tri_out, torch_stand)


try:
    from flash_attn.flash_attn_interface import \
        flash_attn_qkvpacked_func as flash_attn_func
    HAS_FLASH = True
except BaseException:
    HAS_FLASH = False
HAS_FLASH = False
TORCH_HAS_FP8 = hasattr(torch, 'float8_e4m3fn')
BATCH, N_HEADS, HEAD_DIM = 4, 32, 64
configs = []
for mode in ["fwd"]:  
    for causal in [False]:  
        if mode == "bwd" and not causal:
            continue
        configs.append(
            triton.testing.Benchmark(
                x_names=["N_CTX"],
                x_vals=[2**i for i in range(10, 16)],
                line_arg="provider",
                line_vals=(["triton-fp8"] if TORCH_HAS_FP8 else []) +
                (["flash"] if HAS_FLASH else []),
                line_names=(["Triton [FP8]"] if TORCH_HAS_FP8 else []) +
                (["Flash-2"] if HAS_FLASH else []),
                styles=[("red", "-"), ("blue", "-")],
                ylabel="ms",
                plot_name=f"fused-attention-batch{BATCH}-head{N_HEADS}-d{HEAD_DIM}-{mode}-causal={causal}",
                args={
                    "H": N_HEADS,
                    "BATCH": BATCH,
                    "HEAD_DIM": HEAD_DIM,
                    "mode": mode,
                    "causal": causal,
                },
            ))


@triton.testing.perf_report(configs)
def bench_flash_attention(BATCH, H, N_CTX, HEAD_DIM, causal, mode, provider, device="cuda"):
    assert mode in ["fwd", "bwd"]
    warmup = 25
    rep = 100
    dtype = torch.float16
    if "triton" in provider:
        q = torch.randn((BATCH, H, N_CTX, HEAD_DIM), dtype=dtype, device=device, requires_grad=True).contiguous()
        k = torch.randn((BATCH, H, N_CTX, HEAD_DIM), dtype=dtype, device=device, requires_grad=True).contiguous()
        v = torch.randn((BATCH, H, N_CTX, HEAD_DIM), dtype=dtype, device=device, requires_grad=True).contiguous()
        q_int8, q_scale = per_block_int8(q, BLOCK_M=128)
        k_int8, k_scale = per_block_int8(k, BLOCK_M=64)
        q_int8 = q_int8.to(torch.int8)
        k_int8 = k_int8.to(torch.int8)
        sm_scale = HEAD_DIM**-0.5
        fn = lambda: attention(q_int8, k_int8, v, q_scale, k_scale, causal, sm_scale)
        ms = triton.testing.do_bench(fn, warmup=warmup, rep=rep)
    flops_per_matmul = 2.0 * BATCH * H * N_CTX * N_CTX * HEAD_DIM
    total_flops = 2 * flops_per_matmul
    if causal:
        total_flops *= 0.5
    return total_flops / ms * 1e-9


if __name__ == "__main__":
    bench_flash_attention.run(print_data=True)
    test_op()
