import torch

import triton
import triton.language as tl

@triton.jit
def q_kernel_per_token_int8(X, X_int8, BLK: tl.constexpr, Scale, L, C: tl.constexpr):
    off_b = tl.program_id(1) 
    off_blk = tl.program_id(0)
    x_offset = off_b * L * C 
    offs_m = off_blk*BLK + tl.arange(0, BLK)
    offs_k = tl.arange(0, C)

    x_ptrs = X + x_offset + offs_m[:, None] * C + offs_k[None, :]
    x_int8_ptrs = X_int8 + x_offset + offs_m[:, None] * C + offs_k[None, :]
    scale_ptrs = Scale + x_offset//C + offs_m[:, None]  

    x = tl.load(x_ptrs, mask=offs_m[:, None] < L)
    scale = tl.max(tl.abs(x), axis=1, keep_dims=True) / 127.
    x_int8 = x / scale
    x_int8 += 0.5 * tl.where(x_int8 >= 0, 1, -1)
    x_int8 = x_int8.to(tl.int8)
    tl.store(x_int8_ptrs, x_int8, mask=offs_m[:, None] < L)
    tl.store(scale_ptrs, scale, mask=offs_m[:, None] < L)

@triton.jit
def k_kernel_per_token_int8(X, X_int8, BLK: tl.constexpr, Scale, L, C: tl.constexpr):
    off_b = tl.program_id(1) 
    off_blk = tl.program_id(0)
    x_offset = off_b * L * C 
    offs_m = off_blk*BLK + tl.arange(0, BLK)
    offs_k = tl.arange(0, C)

    x_ptrs = X + x_offset + offs_m[:, None] * C + offs_k[None, :]
    x_int8_ptrs = X_int8 + x_offset + offs_m[:, None] * C + offs_k[None, :]
    scale_ptrs = Scale + x_offset//C + offs_m[:, None]  

    x = tl.load(x_ptrs, mask=offs_m[:, None] < L)
    scale = tl.max(tl.abs(x), axis=1, keep_dims=True) / 127.
    x_int8 = x / scale
    x_int8 += 0.5 * tl.where(x_int8 >= 0, 1, -1)
    x_int8 = x_int8.to(tl.int8)
    tl.store(x_int8_ptrs, x_int8, mask=offs_m[:, None] < L)
    tl.store(scale_ptrs, scale, mask=offs_m[:, None] < L)


def per_token_int8(q, k, BLK=128):
    q_int8 = torch.empty_like(q, dtype=torch.int8)
    if q.dim() == 3:
        q_scale = torch.empty((q.shape[-3], q.shape[-2], 1), device=q.device, dtype=torch.float32)
    elif q.dim() == 4:
        q_scale = torch.empty((q.shape[-4], q.shape[-3], q.shape[-2], 1), device=q.device, dtype=torch.float32)
    k_int8 = q_int8.clone()
    k_scale = q_scale.clone()
    q = q.view(-1, q.shape[-2], q.shape[-1])
    k = k.view(-1, k.shape[-2], k.shape[-1])
    B, L, C = q.shape
    grid = ((L+BLK-1)//BLK, B, )
    q_kernel_per_token_int8[grid](
        q, 
        q_int8,
        BLK,
        q_scale,
        L, C,
    )
    k_kernel_per_token_int8[grid](
        k, 
        k_int8,
        BLK,
        k_scale,
        L, C,
    )
    return q_int8, q_scale, k_int8, k_scale
