from einops import rearrange, repeat
import torch
from torch import nn, einsum, broadcast_tensors, Tensor

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

def per_block_int8(tensor, BLOCK_M=128):
    num_blocks =  (tensor.shape[2] // BLOCK_M) + (1 if tensor.shape[2] % BLOCK_M != 0 else 0)
    processed_tensors = []
    scales = []
    
    for i in range(num_blocks):
        start_idx = i * BLOCK_M
        end_idx = min(start_idx + BLOCK_M, tensor.shape[2])
        
        block_tensor = tensor[:, :, start_idx:end_idx, :]
        
        max_over_h, _ = torch.max(torch.abs(block_tensor), dim=2, keepdim=True)

        max_over_hw, _ = torch.max(max_over_h, dim=3, keepdim=True)
        scale = max_over_hw / 127
        
        scaled_tensor = (block_tensor / scale).round().to(torch.int8).to(torch.float32)
        
        processed_tensors.append(scaled_tensor)
        scales.append(scale)
    
    final_tensor = torch.cat(processed_tensors, dim=2)
    final_scales = torch.cat(scales, dim=2)

    return final_tensor, final_scales

def forward(
        seq_len = None,
        head_dim = 64,
        offset = 0,
    ):
        t = ((torch.arange(seq_len) + 0) / 1).to(device)
        theta = 10000
        theta *= 1 ** (head_dim / (head_dim - 2))
        freqs = 1. / (theta ** (torch.arange(0, head_dim, 2)[:(head_dim // 2)].float().to(device) / head_dim))
        freqs = einsum('..., f -> ... f', t.type(freqs.dtype), freqs)
        freqs = repeat(freqs, '... n -> ... (n r)', r = 2)
        return freqs

def rotate_half(x):
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return torch.cat((-x2, x1), dim=-1)


def apply_rotary_emb(freqs, t, start_index = 0, scale = 1., seq_dim = -2):
    dtype = t.dtype

    if t.ndim == 3:
        seq_len = t.shape[seq_dim]
        freqs = freqs[-seq_len:]

    rot_dim = freqs.shape[-1]
    end_index = start_index + rot_dim

    assert rot_dim <= t.shape[-1], f'feature dimension {t.shape[-1]} is not of sufficient size to rotate in all the positions {rot_dim}'

    t_left, t, t_right = t[..., :start_index], t[..., start_index:end_index], t[..., end_index:]
    t = (t * freqs.cos() * scale) + (rotate_half(t) * freqs.sin() * scale)
    out = torch.cat((t_left, t, t_right), dim = -1)

    return out.type(dtype)


def apply_rotary_pos_emb(q, k, cos, sin, position_ids=None, unsqueeze_dim=1):
    
    
    q_embed = (q * cos) + (rotate_half(q) * sin)
    k_embed = (k * cos) + (rotate_half(k) * sin)
    return q_embed, k_embed



import triton
from triton import language as tl

@triton.jit
def _inplace_fused_rope_3d_forward(
    X, XINT8, COS, SIN, SCALE,
    stride_x_b, stride_x_s, stride_s_b,
    seq_len,
    offset,
    ROPE_DIM: tl.constexpr,
    BLOCK_N: tl.constexpr
):
    start_m = tl.program_id(0)
    off_b = tl.program_id(1)

    pos_offs = start_m * BLOCK_N + tl.arange(0, BLOCK_N)[:, None]
    x_ptrs = X + stride_x_b * off_b + pos_offs * stride_x_s + tl.arange(0, ROPE_DIM)[None, :]
    x_int8_ptrs = XINT8 + stride_x_b * off_b + pos_offs * stride_x_s + tl.arange(0, ROPE_DIM)[None, :]
    cos_ptrs = COS + (offset + pos_offs) * ROPE_DIM + tl.arange(0, ROPE_DIM)[None, :]
    sin_ptrs = SIN + (offset + pos_offs) * ROPE_DIM + tl.arange(0, ROPE_DIM)[None, :]
    x = tl.load(x_ptrs, mask = pos_offs < seq_len)
    cos = tl.load(cos_ptrs, mask = pos_offs < seq_len)
    sin = tl.load(sin_ptrs, mask = pos_offs < seq_len)
    index = (tl.arange(0, ROPE_DIM) + ROPE_DIM // 2) % ROPE_DIM
    indicator = tl.arange(0, ROPE_DIM)[None, :] == index[:, None]
    s = tl.where(index < ROPE_DIM//2, 1, -1)
    rot_x = tl.sum(x[:, :, None] * indicator[None, :, :], axis=1) * s[None, :]
    res = x.to(tl.float32) * cos.to(tl.float32) + rot_x.to(tl.float32) * sin.to(tl.float32)
    res *= (ROPE_DIM**-0.5 * 1.44269504)
    scale = tl.max(tl.abs(res)) / 127
    scale_ptrs = SCALE + off_b * stride_s_b + start_m  
    
    res_int8 = res / scale
    res_int8 = res_int8 + 0.5 * tl.where(res_int8 >= 0, 1, -1) 
    res_int8 = res_int8.to(tl.int8)

    tl.store(x_int8_ptrs, res_int8, mask = pos_offs < seq_len)
    tl.store(scale_ptrs, scale)


def rope_quant(x: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor, rope_dim: int, BLK : int=128, rope_offset: int=0):  

    x_int8 = torch.empty_like(x, dtype=torch.int8)
    if x.dim() == 3:
        scale = torch.empty((x.shape[-3], (x.shape[-2] + BLK - 1) // BLK, 1), device=x.device, dtype=torch.float32)
    elif x.dim() == 4:
        scale = torch.empty((x.shape[-4], x.shape[-3], (x.shape[-2] + BLK - 1) // BLK, 1), device=x.device, dtype=torch.float32)
    x = x.view(-1, x.shape[-2], x.shape[-1])
    assert cos.dim() == 2 and sin.dim() == 2
    
    b, s, h = x.shape
    assert cos.size(0) >= s + rope_offset and cos.size(1) == rope_dim
    assert sin.size(0) >= s + rope_offset and sin.size(1) == rope_dim

    BLOCK_N = BLK
    grid = (triton.cdiv(s, BLOCK_N), b)
    _inplace_fused_rope_3d_forward[grid](
        x, x_int8, cos, sin, scale, x.stride(0), x.stride(1), scale.stride(1),   
        s, offset=rope_offset, ROPE_DIM=rope_dim, BLOCK_N = BLOCK_N
    )

    return x_int8, scale
