import os

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"

import torch
import torch.nn as nn
from transformers.models.llama.modeling_llama import (
    LlamaAttention,
    LlamaDecoderLayer,
    LlamaForCausalLM,
    LlamaMLP,
)
from transformers import LlamaTokenizer
import tqdm
import argparse
LENGTH = 2048
class Evaluator:
    def __init__(self, dataset, tokenizer, device, n_samples=40):
        self.dataset = dataset
        self.tokenizer = tokenizer
        self.device = device

        self.dataset = tokenizer(
            "\n\n".join(dataset["text"]), return_tensors="pt"
        ).input_ids.to(device)

        self.n_samples = n_samples

    @torch.no_grad()
    def evaluate(self, model, kernel_args):
        model.eval()
        nlls = []
        
        for i in tqdm.tqdm(range(self.n_samples), desc="Evaluating..."):
            batch = self.dataset[:, (i * LENGTH) : ((i + 1) * LENGTH)].to(model.device)
            with torch.no_grad():
                lm_logits = model(kernel_args, batch).logits
            shift_logits = lm_logits[:, :-1, :].contiguous().float()
            shift_labels = self.dataset[:, (i * LENGTH) : ((i + 1) * LENGTH)][:, 1:]
            loss_fct = nn.CrossEntropyLoss()
            loss = loss_fct(
                shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1)
            )
            neg_log_likelihood = loss.float() * LENGTH
            nlls.append(neg_log_likelihood)

        return torch.exp(torch.stack(nlls).sum() / (self.n_samples * LENGTH))


from datasets import load_dataset

py_parser = argparse.ArgumentParser(add_help=False)
py_parser.add_argument("--quant_v", "-v", action='store_true', help="whether to quant v")
py_parser.add_argument("--minusk", "-k", action='store_true', help="whether to use k-k.mean()")
py_parser.add_argument("--type", type=str, default='block', help="which quant to use")
py_parser.add_argument("--load_max", action='store_true', help="load max from memory")
py_parser.add_argument("--attn_kernel_name", "-name", type=str, default='attn_qk_int8_per_block_hd128_causal', help="which kernel to use")
kernel_args, _ = py_parser.parse_known_args()
model_dir = "./llama2-7b-hf-model"
tokenizer = LlamaTokenizer.from_pretrained(model_dir)
dataset = load_dataset('wikitext', 'wikitext-2-raw-v1', split='test')
evaluator = Evaluator(dataset, tokenizer, "cuda:0")

model_fp16 = LlamaForCausalLM.from_pretrained(
    model_dir, torch_dtype=torch.float16, device_map="cuda", attn_implementation = "sdpa"
)
ppl_fp16 = evaluator.evaluate(model_fp16, kernel_args)
print(f"Original model (fp16) perplexity: {ppl_fp16}")