

import unittest

import numpy as np

from transformers.testing_utils import require_torch, require_vision
from transformers.utils import is_torch_available, is_vision_available

from ...test_image_processing_common import ImageProcessingTestMixin, prepare_video_inputs


if is_torch_available():
    import torch

if is_vision_available():
    from PIL import Image

    from transformers import VivitImageProcessor


class VivitImageProcessingTester(unittest.TestCase):
    def __init__(
        self,
        parent,
        batch_size=7,
        num_channels=3,
        num_frames=10,
        image_size=18,
        min_resolution=30,
        max_resolution=400,
        do_resize=True,
        size=None,
        do_normalize=True,
        image_mean=[0.5, 0.5, 0.5],
        image_std=[0.5, 0.5, 0.5],
        crop_size=None,
    ):
        super().__init__()
        size = size if size is not None else {"shortest_edge": 18}
        crop_size = crop_size if crop_size is not None else {"height": 18, "width": 18}

        self.parent = parent
        self.batch_size = batch_size
        self.num_channels = num_channels
        self.num_frames = num_frames
        self.image_size = image_size
        self.min_resolution = min_resolution
        self.max_resolution = max_resolution
        self.do_resize = do_resize
        self.size = size
        self.do_normalize = do_normalize
        self.image_mean = image_mean
        self.image_std = image_std
        self.crop_size = crop_size

    def prepare_image_processor_dict(self):
        return {
            "image_mean": self.image_mean,
            "image_std": self.image_std,
            "do_normalize": self.do_normalize,
            "do_resize": self.do_resize,
            "size": self.size,
            "crop_size": self.crop_size,
        }

    def expected_output_image_shape(self, images):
        return self.num_frames, self.num_channels, self.crop_size["height"], self.crop_size["width"]

    def prepare_video_inputs(self, equal_resolution=False, numpify=False, torchify=False):
        return prepare_video_inputs(
            batch_size=self.batch_size,
            num_channels=self.num_channels,
            num_frames=self.num_frames,
            min_resolution=self.min_resolution,
            max_resolution=self.max_resolution,
            equal_resolution=equal_resolution,
            numpify=numpify,
            torchify=torchify,
        )


@require_torch
@require_vision
class VivitImageProcessingTest(ImageProcessingTestMixin, unittest.TestCase):
    image_processing_class = VivitImageProcessor if is_vision_available() else None

    def setUp(self):
        super().setUp()
        self.image_processor_tester = VivitImageProcessingTester(self)

    @property
    def image_processor_dict(self):
        return self.image_processor_tester.prepare_image_processor_dict()

    def test_image_processor_properties(self):
        image_processing = self.image_processing_class(**self.image_processor_dict)
        self.assertTrue(hasattr(image_processing, "image_mean"))
        self.assertTrue(hasattr(image_processing, "image_std"))
        self.assertTrue(hasattr(image_processing, "do_normalize"))
        self.assertTrue(hasattr(image_processing, "do_resize"))
        self.assertTrue(hasattr(image_processing, "do_center_crop"))
        self.assertTrue(hasattr(image_processing, "size"))

    def test_image_processor_from_dict_with_kwargs(self):
        image_processor = self.image_processing_class.from_dict(self.image_processor_dict)
        self.assertEqual(image_processor.size, {"shortest_edge": 18})
        self.assertEqual(image_processor.crop_size, {"height": 18, "width": 18})

        image_processor = self.image_processing_class.from_dict(self.image_processor_dict, size=42, crop_size=84)
        self.assertEqual(image_processor.size, {"shortest_edge": 42})
        self.assertEqual(image_processor.crop_size, {"height": 84, "width": 84})

    def test_rescale(self):
                image = np.arange(0, 256, 1, dtype=np.uint8).reshape(1, 8, 32)

        image_processor = self.image_processing_class(**self.image_processor_dict)

        rescaled_image = image_processor.rescale(image, scale=1 / 127.5)
        expected_image = (image * (1 / 127.5)).astype(np.float32) - 1
        self.assertTrue(np.allclose(rescaled_image, expected_image))

        rescaled_image = image_processor.rescale(image, scale=1 / 255, offset=False)
        expected_image = (image / 255.0).astype(np.float32)
        self.assertTrue(np.allclose(rescaled_image, expected_image))

    def test_call_pil(self):
                image_processing = self.image_processing_class(**self.image_processor_dict)
                video_inputs = self.image_processor_tester.prepare_video_inputs(equal_resolution=False)
        for video in video_inputs:
            self.assertIsInstance(video, list)
            self.assertIsInstance(video[0], Image.Image)

                encoded_videos = image_processing(video_inputs[0], return_tensors="pt").pixel_values
        expected_output_video_shape = self.image_processor_tester.expected_output_image_shape([encoded_videos[0]])
        self.assertEqual(tuple(encoded_videos.shape), (1, *expected_output_video_shape))

                encoded_videos = image_processing(video_inputs, return_tensors="pt").pixel_values
        expected_output_video_shape = self.image_processor_tester.expected_output_image_shape(encoded_videos)
        self.assertEqual(
            tuple(encoded_videos.shape), (self.image_processor_tester.batch_size, *expected_output_video_shape)
        )

    def test_call_numpy(self):
                image_processing = self.image_processing_class(**self.image_processor_dict)
                video_inputs = self.image_processor_tester.prepare_video_inputs(equal_resolution=False, numpify=True)
        for video in video_inputs:
            self.assertIsInstance(video, list)
            self.assertIsInstance(video[0], np.ndarray)

                encoded_videos = image_processing(video_inputs[0], return_tensors="pt").pixel_values
        expected_output_video_shape = self.image_processor_tester.expected_output_image_shape([encoded_videos[0]])
        self.assertEqual(tuple(encoded_videos.shape), (1, *expected_output_video_shape))

                encoded_videos = image_processing(video_inputs, return_tensors="pt").pixel_values
        expected_output_video_shape = self.image_processor_tester.expected_output_image_shape(encoded_videos)
        self.assertEqual(
            tuple(encoded_videos.shape), (self.image_processor_tester.batch_size, *expected_output_video_shape)
        )

    def test_call_numpy_4_channels(self):
                image_processing = self.image_processing_class(**self.image_processor_dict)
                self.image_processor_tester.num_channels = 4
        video_inputs = self.image_processor_tester.prepare_video_inputs(equal_resolution=False, numpify=True)
        for video in video_inputs:
            self.assertIsInstance(video, list)
            self.assertIsInstance(video[0], np.ndarray)

                encoded_videos = image_processing(
            video_inputs[0], return_tensors="pt", image_mean=0, image_std=1, input_data_format="channels_first"
        ).pixel_values
        expected_output_video_shape = self.image_processor_tester.expected_output_image_shape([encoded_videos[0]])
        self.assertEqual(tuple(encoded_videos.shape), (1, *expected_output_video_shape))

                encoded_videos = image_processing(
            video_inputs, return_tensors="pt", image_mean=0, image_std=1, input_data_format="channels_first"
        ).pixel_values
        expected_output_video_shape = self.image_processor_tester.expected_output_image_shape(encoded_videos)
        self.assertEqual(
            tuple(encoded_videos.shape), (self.image_processor_tester.batch_size, *expected_output_video_shape)
        )
        self.image_processor_tester.num_channels = 3

    def test_call_pytorch(self):
                image_processing = self.image_processing_class(**self.image_processor_dict)
                video_inputs = self.image_processor_tester.prepare_video_inputs(equal_resolution=False, torchify=True)
        for video in video_inputs:
            self.assertIsInstance(video, list)
            self.assertIsInstance(video[0], torch.Tensor)

                encoded_videos = image_processing(video_inputs[0], return_tensors="pt").pixel_values
        expected_output_video_shape = self.image_processor_tester.expected_output_image_shape([encoded_videos[0]])
        self.assertEqual(tuple(encoded_videos.shape), (1, *expected_output_video_shape))

                encoded_videos = image_processing(video_inputs, return_tensors="pt").pixel_values
        expected_output_video_shape = self.image_processor_tester.expected_output_image_shape(encoded_videos)
        self.assertEqual(
            tuple(encoded_videos.shape), (self.image_processor_tester.batch_size, *expected_output_video_shape)
        )
