















import sys
import tempfile
import unittest
import unittest.mock as mock
from pathlib import Path

from huggingface_hub import HfFolder, delete_repo
from requests.exceptions import HTTPError

from transformers import AutoFeatureExtractor, Wav2Vec2FeatureExtractor
from transformers.testing_utils import TOKEN, USER, get_tests_dir, is_staging_test


sys.path.append(str(Path(__file__).parent.parent.parent / "utils"))

from test_module.custom_feature_extraction import CustomFeatureExtractor  


SAMPLE_FEATURE_EXTRACTION_CONFIG_DIR = get_tests_dir("fixtures")


class FeatureExtractorUtilTester(unittest.TestCase):
    def test_cached_files_are_used_when_internet_is_down(self):
        
        response_mock = mock.Mock()
        response_mock.status_code = 500
        response_mock.headers = {}
        response_mock.raise_for_status.side_effect = HTTPError
        response_mock.json.return_value = {}

        
        _ = Wav2Vec2FeatureExtractor.from_pretrained("hf-internal-testing/tiny-random-wav2vec2")
        
        with mock.patch("requests.Session.request", return_value=response_mock) as mock_head:
            _ = Wav2Vec2FeatureExtractor.from_pretrained("hf-internal-testing/tiny-random-wav2vec2")
            
            mock_head.assert_called()


@is_staging_test
class FeatureExtractorPushToHubTester(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._token = TOKEN
        HfFolder.save_token(TOKEN)

    @staticmethod
    def _try_delete_repo(repo_id, token):
        try:
            
            delete_repo(repo_id=repo_id, token=token)
        except:  
            pass

    def test_push_to_hub(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            try:
                tmp_repo = f"{USER}/test-feature-extractor-{Path(tmp_dir).name}"

                feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained(SAMPLE_FEATURE_EXTRACTION_CONFIG_DIR)
                feature_extractor.push_to_hub(tmp_repo, token=self._token)

                new_feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained(tmp_repo)
                for k, v in feature_extractor.__dict__.items():
                    self.assertEqual(v, getattr(new_feature_extractor, k))
            finally:
                
                self._try_delete_repo(repo_id=tmp_repo, token=self._token)

    def test_push_to_hub_via_save_pretrained(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            try:
                tmp_repo = f"{USER}/test-feature-extractor-{Path(tmp_dir).name}"
                feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained(SAMPLE_FEATURE_EXTRACTION_CONFIG_DIR)
                
                feature_extractor.save_pretrained(tmp_dir, repo_id=tmp_repo, push_to_hub=True, token=self._token)

                new_feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained(tmp_repo)
                for k, v in feature_extractor.__dict__.items():
                    self.assertEqual(v, getattr(new_feature_extractor, k))
            finally:
                
                self._try_delete_repo(repo_id=tmp_repo, token=self._token)

    def test_push_to_hub_in_organization(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            try:
                tmp_repo = f"valid_org/test-feature-extractor-{Path(tmp_dir).name}"
                feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained(SAMPLE_FEATURE_EXTRACTION_CONFIG_DIR)
                feature_extractor.push_to_hub(tmp_repo, token=self._token)

                new_feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained(tmp_repo)
                for k, v in feature_extractor.__dict__.items():
                    self.assertEqual(v, getattr(new_feature_extractor, k))
            finally:
                
                self._try_delete_repo(repo_id=tmp_repo, token=self._token)

    def test_push_to_hub_in_organization_via_save_pretrained(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            try:
                tmp_repo = f"valid_org/test-feature-extractor-{Path(tmp_dir).name}"
                feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained(SAMPLE_FEATURE_EXTRACTION_CONFIG_DIR)
                
                feature_extractor.save_pretrained(tmp_dir, repo_id=tmp_repo, push_to_hub=True, token=self._token)

                new_feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained(tmp_repo)
                for k, v in feature_extractor.__dict__.items():
                    self.assertEqual(v, getattr(new_feature_extractor, k))
            finally:
                
                self._try_delete_repo(repo_id=tmp_repo, token=self._token)

    def test_push_to_hub_dynamic_feature_extractor(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            try:
                tmp_repo = f"{USER}/test-dynamic-feature-extractor-{Path(tmp_dir).name}"
                CustomFeatureExtractor.register_for_auto_class()
                feature_extractor = CustomFeatureExtractor.from_pretrained(SAMPLE_FEATURE_EXTRACTION_CONFIG_DIR)

                feature_extractor.push_to_hub(tmp_repo, token=self._token)

                
                self.assertDictEqual(
                    feature_extractor.auto_map,
                    {"AutoFeatureExtractor": "custom_feature_extraction.CustomFeatureExtractor"},
                )

                new_feature_extractor = AutoFeatureExtractor.from_pretrained(tmp_repo, trust_remote_code=True)
                
                self.assertEqual(new_feature_extractor.__class__.__name__, "CustomFeatureExtractor")
            finally:
                
                self._try_delete_repo(repo_id=tmp_repo, token=self._token)
