
import contextlib
import importlib
import io
import unittest

import transformers

from transformers import *  from transformers.testing_utils import DUMMY_UNKNOWN_IDENTIFIER, require_flax, require_tf, require_torch
from transformers.utils import ContextManagers, find_labels, is_flax_available, is_tf_available, is_torch_available


if is_torch_available():
    from transformers import BertForPreTraining, BertForQuestionAnswering, BertForSequenceClassification

if is_tf_available():
    from transformers import TFBertForPreTraining, TFBertForQuestionAnswering, TFBertForSequenceClassification

if is_flax_available():
    from transformers import FlaxBertForPreTraining, FlaxBertForQuestionAnswering, FlaxBertForSequenceClassification


MODEL_ID = DUMMY_UNKNOWN_IDENTIFIER

REVISION_ID_DEFAULT = "main"
REVISION_ID_ONE_SPECIFIC_COMMIT = "f2c752cfc5c0ab6f4bdec59acea69eefbee381c2"
REVISION_ID_INVALID = "aaaaaaa"

PINNED_SHA1 = "d9e9f15bc825e4b2c9249e9578f884bbcb5e3684"
PINNED_SHA256 = "4b243c475af8d0a7754e87d7d096c92e5199ec2fe168a2ee7998e3b8e9bcb1d3"


@contextlib.contextmanager
def context_en():
    print("Welcome!")
    yield
    print("Bye!")


@contextlib.contextmanager
def context_fr():
    print("Bonjour!")
    yield
    print("Au revoir!")


class TestImportMechanisms(unittest.TestCase):
    def test_module_spec_available(self):
                assert transformers.__spec__ is not None
        assert importlib.util.find_spec("transformers") is not None


class GenericUtilTests(unittest.TestCase):
    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_context_managers_no_context(self, mock_stdout):
        with ContextManagers([]):
            print("Transformers are awesome!")
                self.assertEqual(mock_stdout.getvalue(), "Transformers are awesome!\n")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_context_managers_one_context(self, mock_stdout):
        with ContextManagers([context_en()]):
            print("Transformers are awesome!")
                self.assertEqual(mock_stdout.getvalue(), "Welcome!\nTransformers are awesome!\nBye!\n")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_context_managers_two_context(self, mock_stdout):
        with ContextManagers([context_fr(), context_en()]):
            print("Transformers are awesome!")
                self.assertEqual(mock_stdout.getvalue(), "Bonjour!\nWelcome!\nTransformers are awesome!\nBye!\nAu revoir!\n")

    @require_torch
    def test_find_labels_pt(self):
        self.assertEqual(find_labels(BertForSequenceClassification), ["labels"])
        self.assertEqual(find_labels(BertForPreTraining), ["labels", "next_sentence_label"])
        self.assertEqual(find_labels(BertForQuestionAnswering), ["start_positions", "end_positions"])

                class DummyModel(BertForSequenceClassification):
            pass

        self.assertEqual(find_labels(DummyModel), ["labels"])

    @require_tf
    def test_find_labels_tf(self):
        self.assertEqual(find_labels(TFBertForSequenceClassification), ["labels"])
        self.assertEqual(find_labels(TFBertForPreTraining), ["labels", "next_sentence_label"])
        self.assertEqual(find_labels(TFBertForQuestionAnswering), ["start_positions", "end_positions"])

                class DummyModel(TFBertForSequenceClassification):
            pass

        self.assertEqual(find_labels(DummyModel), ["labels"])

    @require_flax
    def test_find_labels_flax(self):
                self.assertEqual(find_labels(FlaxBertForSequenceClassification), [])
        self.assertEqual(find_labels(FlaxBertForPreTraining), [])
        self.assertEqual(find_labels(FlaxBertForQuestionAnswering), [])

                class DummyModel(FlaxBertForSequenceClassification):
            pass

        self.assertEqual(find_labels(DummyModel), [])
