"""
This should help you prepare a patch, automatically extracting the commits to cherry-pick
in chronological order to avoid merge conflicts. An equivalent way to do this is to use
`git log --pretty=oneline HEAD...v4.41.0` and grep.

Potential TODO: automatically cherry-picks them.

Pass in a list of PR:
`python utils/patch_helper.py --prs 31108 31054 31008 31010 31004`
will produce the following:
```bash
Skipping invalid version tag: list
Skipping invalid version tag: localattn1
Git cherry-pick commands to run:
git cherry-pick 03935d300d60110bb86edb49d2315089cfb19789 git cherry-pick bdb9106f247fca48a71eb384be25dbbd29b065a8 git cherry-pick 84c4b72ee99e8e65a8a5754a5f9d6265b45cf67e git cherry-pick 936ab7bae5e040ec58994cb722dd587b9ab26581 git cherry-pick 0bef4a273825d2cfc52ddfe62ba486ee61cc116f ```
"""

import argparse

from git import GitCommandError, Repo
from packaging import version


def get_merge_commit(repo, pr_number, since_tag):
    try:
                merge_commit = next(repo.iter_commits(f"v{since_tag}...HEAD", grep=f"        return merge_commit
    except StopIteration:
        print(f"No merge commit found for PR         return None
    except GitCommandError as e:
        print(f"Error finding merge commit for PR         return None


def main(pr_numbers):
    repo = Repo(".")      merge_commits = []

    tags = {}
    for tag in repo.tags:
        try:
                        tag_ver = version.parse(tag.name)
            tags[tag_ver] = tag
        except Exception:
            print(f"Skipping invalid version tag: {tag.name}")

    last_tag = sorted(tags)[-1]
    major_minor = f"{last_tag.major}.{last_tag.minor}.0"
        for pr in pr_numbers:
        commit = get_merge_commit(repo, pr, major_minor)
        if commit:
            merge_commits.append(commit)

        merge_commits.sort(key=lambda commit: commit.committed_datetime)

        print("Git cherry-pick commands to run:")
    for commit in merge_commits:
        print(f"git cherry-pick {commit.hexsha} 

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Find and sort merge commits for specified PRs.")
    parser.add_argument("--prs", nargs="+", required=True, type=int, help="PR numbers to find merge commits for")

    args = parser.parse_args()
    main(args.prs)
