import torch, pickle, importlib
import torch.nn as nn
import torch.nn.functional as F
import math
from einops import rearrange
from modules.speed_util import checkpoint
import sys
sys.path.append("../../Kernels/")
from quant_per_token import per_token_int8
from quant_per_block import per_block_int8
from quant_v import perchannel_absmax_quantize

sys.path.append("../")
from utils_infer_attn import precision_metric

class Linear(torch.nn.Linear):
    def reset_parameters(self):
        return None

class Conv2d(torch.nn.Conv2d):
    def reset_parameters(self):
        return None

class AttnBlock_lrfuse_backup(nn.Module):
    def __init__(self, c, c_cond, nhead, self_attn=True, dropout=0.0, use_checkpoint=True):
        super().__init__()
        self.self_attn = self_attn
        self.norm = LayerNorm2d(c, elementwise_affine=False, eps=1e-6)
        self.attention = Attention2D(c, nhead, dropout)
        self.kv_mapper = nn.Sequential(
            nn.SiLU(),
            Linear(c_cond, c)
        )
        self.fuse_mapper = nn.Sequential(
            nn.SiLU(),
            Linear(c_cond, c)
        )
        self.use_checkpoint = use_checkpoint
       
    def forward(self, hr, lr):
        return checkpoint(self._forward, (hr, lr), self.paramters(), self.use_checkpoint)
    def _forward(self, hr, lr):
        res = hr
        hr = self.kv_mapper(rearrange(hr, 'b c h w -> b (h w ) c'))
        lr_fuse = self.attention(self.norm(lr), hr, self_attn=False) + lr

        lr_fuse = self.fuse_mapper(rearrange(lr_fuse, 'b c h w -> b (h w ) c'))
        hr = self.attention(self.norm(res), lr_fuse, self_attn=False) + res
        return hr
        
        
class AttnBlock_lrfuse(nn.Module):
    def __init__(self, c, c_cond, nhead, self_attn=True, dropout=0.0, kernel_size=3, use_checkpoint=True):
        super().__init__()
        self.self_attn = self_attn
        self.norm = LayerNorm2d(c, elementwise_affine=False, eps=1e-6)
        self.attention = Attention2D(c, nhead, dropout)
        self.kv_mapper = nn.Sequential(
            nn.SiLU(),
            Linear(c_cond, c)
        )
      
       
        self.depthwise = Conv2d(c, c , kernel_size=kernel_size, padding=kernel_size // 2, groups=c)
             
        self.channelwise = nn.Sequential(
                    Linear(c + c, c ),
                    nn.GELU(),
                    GlobalResponseNorm(c ),
                    nn.Dropout(dropout),
                    Linear(c , c)
                )
        self.use_checkpoint = use_checkpoint
    
    
    def forward(self, hr, lr):
        return checkpoint(self._forward, (hr, lr), self.parameters(), self.use_checkpoint)
        
    def _forward(self, hr, lr):
        res = hr
        hr = self.kv_mapper(rearrange(hr, 'b c h w -> b (h w ) c'))
        lr_fuse = self.attention(self.norm(lr), hr, self_attn=False) + lr
        
        lr_fuse = torch.nn.functional.interpolate(lr_fuse.float(), res.shape[2:])
                media = torch.cat((self.depthwise(lr_fuse), res), dim=1)
        out = self.channelwise(media.permute(0,2,3,1)).permute(0,3,1,2) + res
        
        return out        
        
        


class Attention2D(nn.Module):  
    def __init__(self, kernel_args, c, nhead, layer_idx, dropout=0.0):
        super().__init__()
        self.attn = nn.MultiheadAttention(c, nhead, dropout=dropout, bias=True, batch_first=True)
        self.nhead = nhead
        self.c = c // self.nhead
        self.layer_idx = layer_idx
        self.kernel_args = kernel_args
        
        

    def forward(self, x, kv, self_attn=False):
        kernel_args = self.kernel_args
        attn_kernel_name = kernel_args.attn_kernel_name
        quant_type = kernel_args.type
        minusk = kernel_args.minusk
        quant_v = kernel_args.quant_v
        load_max = kernel_args.load_max

        attn_kernel = importlib.import_module(attn_kernel_name)

        orig_shape = x.shape
        
        x = x.view(x.size(0), x.size(1), -1).permute(0, 2, 1)  
        if self_attn:
            kv = torch.cat([x, kv], dim=1)

        l, s = x.shape[-2], kv.shape[-2]
        
        if x.shape[1]>1000:
            x = F.pad(x, (0, 0, 0, s-l, 0, 0))

            in_proj_weight = self.attn.in_proj_weight
            in_proj_bias = self.attn.in_proj_bias
            out_proj_weight = self.attn.out_proj.weight
            out_proj_bias = self.attn.out_proj.bias
            
            Bq, Bk, Bv = torch.chunk(in_proj_bias, 3, dim=0)
            Wq, Wk, Wv = torch.chunk(in_proj_weight, 3, dim=0)
            Wo = out_proj_weight
            

            batch_size, seq_len, num_heads, head_dim = x.size(0), x.size(1), self.nhead, self.c

            Q = F.linear(x.transpose(1, 0), Wq, Bq) 
            K = F.linear(kv.transpose(1, 0), Wk, Bk) 
            V = F.linear(kv.transpose(1, 0), Wv, Bv) 
            proj = torch.cat((Q, K, V), dim=2)
            proj = proj.unflatten(-1, (3, x.size(-1))).unsqueeze(0).transpose(0, -2).squeeze(-2).contiguous()
            q, k, v =proj[0], proj[1], proj[2]
            q = q.view(seq_len, batch_size * num_heads, head_dim).transpose(0, 1)
            k = k.view(seq_len, batch_size * num_heads, head_dim).transpose(0, 1)
            v = v.view(seq_len, batch_size * num_heads, head_dim).transpose(0, 1)
            q = q.view(batch_size, num_heads, seq_len, head_dim).contiguous().to(torch.float16)
            k = k.view(batch_size, num_heads, seq_len, head_dim).contiguous().to(torch.float16)
            v = v.view(batch_size, num_heads, seq_len, head_dim).contiguous().to(torch.float16)


            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            

            
            if minusk:
                k -= k.mean(dim=-2, keepdim=True)

            if quant_type == "token":
                query_layer_int8, query_scale, key_layer_int8, key_scale = per_token_int8(q, k)
            elif quant_type == "block":
                query_layer_int8, query_scale, key_layer_int8, key_scale = per_block_int8(q, k)
            if quant_v:
                value_layer_int8, value_scale = perchannel_absmax_quantize(v)
                attn_output_int8 = attn_kernel.forward(query_layer_int8, key_layer_int8, value_layer_int8, query_scale, key_scale, value_scale)
            else:

                if load_max:
                    with open(f'./modules/max_values/q_max/{self.layer_idx}_q_max.pkl', 'rb') as f:
                        q_max = pickle.load(f).cuda().contiguous()
                    with open(f'./modules/max_values/k_max/{self.layer_idx}_k_max.pkl', 'rb') as f:
                        k_max = pickle.load(f).cuda().contiguous()
                    attn_output_int8 = attn_kernel.forward(query_layer_int8, key_layer_int8, v, query_scale, key_scale, q_max, k_max)
                else:
                    attn_output_int8 = attn_kernel.forward(query_layer_int8, key_layer_int8, v, query_scale, key_scale)

            


            


            attn_output = attn_output_int8.permute(2, 0, 1, 3).contiguous().view(batch_size * seq_len, num_heads*head_dim)
            attn_output = F.linear(attn_output, out_proj_weight, out_proj_bias)
            attn_output = attn_output.view(seq_len, batch_size, attn_output.size(1))
            x = attn_output.transpose(1, 0)

            x = x[:,:l,:].to(torch.float16)

        
        
        else:
            x = self.attn(x, kv, kv, need_weights=False)[0]  
        

        x = x.permute(0, 2, 1).view(*orig_shape)  
        return x
class Attention2D_splitpatch(nn.Module):
    def __init__(self, c, nhead, dropout=0.0):
        super().__init__()
        self.attn = nn.MultiheadAttention(c, nhead, dropout=dropout, bias=True, batch_first=True)

    def forward(self, x, kv, self_attn=False):
        orig_shape = x.shape
        
                x = rearrange(x, 'b c (nh wh) (nw ww) -> (b nh nw) (wh ww) c', wh=24, ww=24, nh=orig_shape[-2] // 24, nw=orig_shape[-1] // 24,)
                        if self_attn:
                        num = (orig_shape[-2] // 24) * (orig_shape[-1] // 24)
            kv = torch.cat([x, kv.repeat(num, 1, 1)], dim=1)
                        
           
        x = self.attn(x, kv, kv, need_weights=False)[0]
        x = rearrange(x, ' (b nh nw) (wh ww) c -> b c (nh wh) (nw ww)', b=orig_shape[0], wh=24, ww=24, nh=orig_shape[-2] // 24, nw=orig_shape[-1] // 24)
                
        return x
class Attention2D_extra(nn.Module):
    def __init__(self, c, nhead, dropout=0.0):
        super().__init__()
        self.attn = nn.MultiheadAttention(c, nhead, dropout=dropout, bias=True, batch_first=True)

    def forward(self, x, kv, extra_emb=None, self_attn=False):
        orig_shape = x.shape
        x = x.view(x.size(0), x.size(1), -1).permute(0, 2, 1)  
        num_x = x.shape[1]

       
        if extra_emb is not None:
            ori_extra_shape = extra_emb.shape
            extra_emb = extra_emb.view(extra_emb.size(0), extra_emb.size(1), -1).permute(0, 2, 1)
            x = torch.cat((x, extra_emb), dim=1)  
        if self_attn:
                        kv = torch.cat([x, kv], dim=1) 
        x = self.attn(x, kv, kv, need_weights=False)[0]
        img = x[:, :num_x, :].permute(0, 2, 1).view(*orig_shape)
        if extra_emb is not None:
            fix = x[:, num_x:, :].permute(0, 2, 1).view(*ori_extra_shape)
            return img, fix
        else:
            return img
class AttnBlock_extraq(nn.Module):
    def __init__(self, c, c_cond, nhead, self_attn=True, dropout=0.0):
        super().__init__()
        self.self_attn = self_attn
        self.norm = LayerNorm2d(c, elementwise_affine=False, eps=1e-6)
                self.attention = Attention2D_extra(c, nhead, dropout)
        self.kv_mapper = nn.Sequential(
            nn.SiLU(),
            Linear(c_cond, c)
        )
    
    def forward(self, x, kv, extra_emb=None):
                        
        kv = self.kv_mapper(kv)
        if extra_emb is not None:
            res_x, res_extra = self.attention(self.norm(x), kv, extra_emb=self.norm2(extra_emb), self_attn=self.self_attn)
            x = x + res_x
            extra_emb = extra_emb + res_extra
            return x, extra_emb
        else:
            x = x + self.attention(self.norm(x), kv, self_attn=self.self_attn)
            return x
class AttnBlock_latent2ex(nn.Module):
    def __init__(self, c, c_cond, nhead, self_attn=True, dropout=0.0):
        super().__init__()
        self.self_attn = self_attn
        self.norm = LayerNorm2d(c, elementwise_affine=False, eps=1e-6)
        self.attention = Attention2D(c, nhead, dropout)
        self.kv_mapper = nn.Sequential(
            nn.SiLU(),
            Linear(c_cond, c)
        )

    def forward(self, x, kv):
                kv = F.interpolate(kv.float(), x.shape[2:])
        kv = kv.view(kv.size(0), kv.size(1), -1).permute(0, 2, 1)
        kv = self.kv_mapper(kv)
        x = x + self.attention(self.norm(x), kv, self_attn=self.self_attn)
        return x

class LayerNorm2d(nn.LayerNorm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, x):
        return super().forward(x.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
class AttnBlock_crossbranch(nn.Module):
    def __init__(self, attnmodule, c, c_cond, nhead, self_attn=True, dropout=0.0):
        super().__init__()
        self.attn = AttnBlock(c, c_cond, nhead, self_attn, dropout)
                self.attn.load_state_dict(attnmodule.state_dict())
        self.norm1 = LayerNorm2d(c, elementwise_affine=False, eps=1e-6)
       
        self.channelwise1 = nn.Sequential(
            Linear(c *2, c ),
            nn.GELU(),
            GlobalResponseNorm(c ),
            nn.Dropout(dropout),
            Linear(c, c)
        )
        self.channelwise2 = nn.Sequential(
            Linear(c *2, c ),
            nn.GELU(),
            GlobalResponseNorm(c ),
            nn.Dropout(dropout),
            Linear(c, c)
        )
        self.c = c
    def forward(self, x, kv, main_x):
                
        x = self.channelwise1(torch.cat((x, F.interpolate(main_x.float(), x.shape[2:])), dim=1).permute(0, 2, 3, 1)).permute(0, 3, 1, 2) + x
        x = self.attn(x, kv)
        main_x = self.channelwise2(torch.cat((main_x, F.interpolate(x.float(), main_x.shape[2:])), dim=1).permute(0, 2, 3, 1)).permute(0, 3, 1, 2) + main_x
        return main_x, x

class GlobalResponseNorm(nn.Module):
    "from https://github.com/facebookresearch/ConvNeXt-V2/blob/3608f67cc1dae164790c5d0aead7bf2d73d9719b/models/utils.py    def __init__(self, dim):
        super().__init__()
        self.gamma = nn.Parameter(torch.zeros(1, 1, 1, dim))
        self.beta = nn.Parameter(torch.zeros(1, 1, 1, dim))

    def forward(self, x):
        Gx = torch.norm(x, p=2, dim=(1, 2), keepdim=True)
        Nx = Gx / (Gx.mean(dim=-1, keepdim=True) + 1e-6)
        return self.gamma * (x * Nx) + self.beta + x


class ResBlock(nn.Module):
    def __init__(self, c, c_skip=0, kernel_size=3, dropout=0.0, use_checkpoint =True):  
        super().__init__()
        self.depthwise = Conv2d(c, c, kernel_size=kernel_size, padding=kernel_size // 2, groups=c)
        
        self.norm = LayerNorm2d(c, elementwise_affine=False, eps=1e-6)
        self.channelwise = nn.Sequential(
            Linear(c + c_skip, c * 4),
            nn.GELU(),
            GlobalResponseNorm(c * 4),
            nn.Dropout(dropout),
            Linear(c * 4, c)
        )
        self.use_checkpoint = use_checkpoint
    def forward(self, x, x_skip=None):
    
        if x_skip is not None:
            return checkpoint(self._forward_skip, (x, x_skip), self.parameters(), self.use_checkpoint)
        else:
                        return checkpoint(self._forward_woskip, (x, ), self.parameters(), self.use_checkpoint)
                    
    
    
    def _forward_skip(self, x, x_skip):
        x_res = x
        x = self.norm(self.depthwise(x))
        if x_skip is not None:
            x = torch.cat([x, x_skip], dim=1)
        x = self.channelwise(x.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        return x + x_res
    def _forward_woskip(self, x):
        x_res = x
        x = self.norm(self.depthwise(x))
       
        x = self.channelwise(x.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        return x + x_res

class AttnBlock(nn.Module):
    def __init__(self, kernel_args, c, c_cond, nhead, layer_idx, self_attn=True, dropout=0.0, use_checkpoint=True):
        super().__init__()
        self.self_attn = self_attn
        self.norm = LayerNorm2d(c, elementwise_affine=False, eps=1e-6)
        self.attention = Attention2D(kernel_args, c, nhead, layer_idx, dropout)
        self.kv_mapper = nn.Sequential(
            nn.SiLU(),
            Linear(c_cond, c)
        )
        self.use_checkpoint = use_checkpoint
    def forward(self, x, kv):
        return checkpoint(self._forward, (x, kv), self.parameters(), self.use_checkpoint)
    def _forward(self, x, kv):
        kv = self.kv_mapper(kv)
        res = self.attention(self.norm(x), kv, self_attn=self.self_attn)
        
                        x = x + res
     
        return x
class AttnBlock_mytest(nn.Module):
    def __init__(self, c, c_cond, nhead, self_attn=True, dropout=0.0):
        super().__init__()
        self.self_attn = self_attn
        self.norm = LayerNorm2d(c, elementwise_affine=False, eps=1e-6)
        self.attention = Attention2D(c, nhead, dropout)
        self.kv_mapper = nn.Sequential(
            nn.SiLU(),
            nn.Linear(c_cond, c)
        )
 
    def forward(self, x, kv):
        kv = self.kv_mapper(kv)
        x = x + self.attention(self.norm(x), kv, self_attn=self.self_attn)
        return x

class FeedForwardBlock(nn.Module):
    def __init__(self, c, dropout=0.0):
        super().__init__()
        self.norm = LayerNorm2d(c, elementwise_affine=False, eps=1e-6)
        self.channelwise = nn.Sequential(
            Linear(c, c * 4),
            nn.GELU(),
            GlobalResponseNorm(c * 4),
            nn.Dropout(dropout),
            Linear(c * 4, c)
        )

    def forward(self, x):
        x = x + self.channelwise(self.norm(x).permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        return x


class TimestepBlock(nn.Module):
    def __init__(self, c, c_timestep, conds=['sca'], use_checkpoint=True):
        super().__init__()
        self.mapper = Linear(c_timestep, c * 2)
        self.conds = conds
        for cname in conds:
            setattr(self, f"mapper_{cname}", Linear(c_timestep, c * 2))

        self.use_checkpoint = use_checkpoint
    def forward(self, x, t):
        return checkpoint(self._forward, (x, t), self.parameters(), self.use_checkpoint)
        
    def _forward(self, x, t):
                        t = t.chunk(len(self.conds) + 1, dim=1)
        a, b = self.mapper(t[0])[:, :, None, None].chunk(2, dim=1)
        for i, c in enumerate(self.conds):
            ac, bc = getattr(self, f"mapper_{c}")(t[i + 1])[:, :, None, None].chunk(2, dim=1)
            a, b = a + ac, b + bc
        return x * (1 + a) + b
